import subprocess
import os
import shutil
from contextlib import suppress

from typing import List

from syncgit._config import SYNCGIT_CMD_TIMEOUT, SYNCGIT_REPO_DIR_NAME


class RepoInfo:
    def __init__(self, local_name: str, url: str, branch: str) -> None:
        self.local_name = local_name
        self.url = url
        self.branch = branch
        self.dir = os.path.join(SYNCGIT_REPO_DIR_NAME, self.local_name)

    def get_abs_path(self, file_name: str) -> str:
        return os.path.join(self.dir, file_name)


def _exec_cmd(command: List[str]) -> None:
    subprocess.run(command, timeout=SYNCGIT_CMD_TIMEOUT,
                   stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True)


def _exec_cmd_str(command: List[str]) -> str:
    subp = subprocess.run(command, timeout=SYNCGIT_CMD_TIMEOUT,
                          stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, check=True)

    return subp.stdout.decode('utf-8')


def _git(repo_info: RepoInfo) -> List[str]:
    return ["git", "-C", repo_info.dir]


def _create_repo_dir(repo_info: RepoInfo) -> None:
    if os.path.exists(repo_info.dir):
        return

    os.makedirs(repo_info.dir)
    _exec_cmd(_git(repo_info) + ["init"])
    _exec_cmd(_git(repo_info) + ["remote", "add", "origin", repo_info.url])


def commit_hash(repo_info: RepoInfo) -> str:
    return _exec_cmd_str(_git(repo_info) + ["show", "-s", "--format=%H"]).rstrip()


def pull(repo_info: RepoInfo) -> str:
    _create_repo_dir(repo_info)

    with suppress(subprocess.SubprocessError):
        _exec_cmd(_git(repo_info) + ["pull", "origin", repo_info.branch])

    return commit_hash(repo_info)


def changes(repo_info: RepoInfo) -> List[str]:
    return _exec_cmd_str(_git(repo_info) + ["diff", "--name-only", "HEAD^", "HEAD"]).splitlines()


def remove(repo_info: RepoInfo) -> None:
    shutil.rmtree(repo_info.dir)


def remove_all() -> None:
    shutil.rmtree(f"./{SYNCGIT_REPO_DIR_NAME}")



import os

SYNCGIT_CMD_TIMEOUT = int(os.getenv("SYNCGIT_CMD_TIMEOUT", "15"))
SYNCGIT_REPO_DIR_NAME = os.getenv("SYNCGIT_REPO_DIR_NAME", ".repos")
SYNCGIT_DEFAULT_POLL_INTERVAL = int(os.getenv("SYNCGIT_POLL_INTERVAL", "5"))



from typing import Any
from threading import Thread

import time


class ThreadLoop:
    def __init__(self, interval: int, callback: Any) -> None:
        self.__interval = interval
        self.__callback = callback
        self.__running = False
        self.__thread: Thread

    def start(self) -> None:
        self.__callback()
        self.__running = True
        self.__thread = Thread(target=self.__loop)
        self.__thread.setDaemon(True)
        self.__thread.start()

    def stop(self) -> None:
        self.__running = False
        self.__thread.join()

    def set_interval(self, seconds: int) -> None:
        self.__interval = seconds

    def __loop(self) -> None:
        starttime = time.time()

        while self.__running:
            self.__callback()
            time.sleep(self.__interval - ((time.time() - starttime) % self.__interval))



from typing import Dict
from typing import List
from typing import Any
from typing import Callable
from typing import Optional
from types import ModuleType

import os
import json
import yaml

from syncgit import _gitcmd as gitcmd
from syncgit._gitcmd import RepoInfo
from syncgit._threadsafe_dict import ThreadSafeDict
from syncgit._threadloop import ThreadLoop
from syncgit._config import SYNCGIT_DEFAULT_POLL_INTERVAL


class UnknownTypeException(Exception):
    pass


class SyncConfig:
    def __init__(self, name: str, path: str, sync_type: str = "auto") -> None:
        '''
        Parameters
        ----------
            name: str
                Name of the attribute alias
            path: str
                Path to file in the repository
            sync_type : str
                File type. Available options are :code:`"text"` (plain text file, will be converted
                to python string), :code:`"json"`, :code:`"yaml"` (converted to python dict)
                and :code:`"module"` (will be imported the file as python module). :code:`"auto"` to
                automatically detect sync type from file extension

        '''
        self.name = name
        self.file_path = path
        self.type = sync_type

        if sync_type == "auto":
            self.__get_auto_type()

    def __get_auto_type(self) -> None:
        file_ext = os.path.splitext(self.file_path)[-1]

        if file_ext in [".yaml", ".yml"]:
            self.type = "yaml"
        elif file_ext == ".json":
            self.type = "json"
        elif file_ext == ".py":
            self.type = "module"
        else:
            self.type = "text"


class Repo:
    def __init__(self, name: str, url: str, branch: str) -> None:
        '''
        Parameters
        ----------
        name : str
            Unique name for this repository and branch
        url : str
            Repository URL to sync to (can be ssh or https)
        branch : str
            Name of the branch to sync to
        '''
        self.__repo_info: RepoInfo = RepoInfo(name, url, branch)
        self.__commit_hash = ""
        self.__values: ThreadSafeDict = ThreadSafeDict()
        self.__config: List[SyncConfig]
        self.__update_callback: Optional[Callable[[Any, List[SyncConfig]], None]] = None
        self.__thread_loop = ThreadLoop(
            SYNCGIT_DEFAULT_POLL_INTERVAL, self.__update)

    @property
    def commit_hash(self) -> str:
        '''
        The commit hash of latest sync
        '''
        return self.__commit_hash

    def set_poll_interval(self, seconds: int) -> None:
        '''
        Set polling interval

        Parameters
        ----------
        seconds : int
            Amount of time between synchronization (in seconds, default is 5 seconds)
        '''
        self.__thread_loop.set_interval(seconds)

    def set_update_callback(self, callback: Callable[[Any, List[SyncConfig]], None]) -> None:
        '''
        Set callback to be call after synchronization is complete

        Parameters
        ----------
        callback : Callable[[Repo, List[SyncConfig]], None]
            This callback will be called when changes are pushed to the repo and
            the new changes have been synchronized. The callback should accept two
            arguments, the Repo class and List[SyncConfig], config attributes which
            got updated
        '''
        self.__update_callback = callback

    def set_config(self, config: List[SyncConfig]) -> None:
        '''
        Configure attributes to sync

        Parameters
        ----------
        config : List[SyncConfig]
            List of SyncConfig specifying attribute names, file path,
            type of the file (see :py:class:`SyncConfig`). These configs will be available
            as class attributes.
        '''
        self.__config = config

    def start_sync(self) -> None:
        '''
        Start sync to git repository
        '''
        self.__thread_loop.start()

    def stop_sync(self) -> None:
        '''
        Stop sync to git repository
        '''
        self.__thread_loop.stop()

    def __pull(self) -> bool:
        new_commit_hash = gitcmd.pull(self.__repo_info)

        if new_commit_hash == self.commit_hash:
            return False

        self.__commit_hash = new_commit_hash
        return True

    def __update(self) -> None:
        if not self.__pull():
            return

        changed_configs = self.__get_changed_configs()

        files = self.__get_files(changed_configs)

        for config in changed_configs:
            self.__set_value(config, files[config.file_path])

        if self.__update_callback is not None:
            self.__update_callback(self, changed_configs)

    def __set_value(self, config: SyncConfig, str_value: str) -> None:
        if config.type == "text":
            self.__values[config.name] = str_value
        elif config.type == "json":
            self.__values[config.name] = json.loads(str_value)
        elif config.type == "yaml":
            self.__values[config.name] = yaml.safe_load(str_value)
        elif config.type == "module":
            self.__set_module(config.name, str_value)
        else:
            raise UnknownTypeException

    def __set_module(self, name: str, str_src: str) -> None:
        compiled = compile(str_src, '', 'exec')
        module = ModuleType(name)
        exec(compiled, module.__dict__)  # pylint: disable=exec-used
        self.__values[name] = module

    def __get_changed_configs(self) -> List[SyncConfig]:
        changes = gitcmd.changes(self.__repo_info)
        first_update = len(self.__values) == 0
        changed_configs = []

        for config in self.__config:
            if config.file_path not in changes and not first_update:
                continue

            changed_configs.append(config)

        return changed_configs

    def __get_files(self, configs: List[SyncConfig]) -> Dict[str, str]:
        files: Dict[str, str] = {}

        for config in configs:
            file_path_absolute = self.__repo_info.get_abs_path(config.file_path)

            with open(file_path_absolute) as file:
                files[config.file_path] = file.read()

        return files

    def __getattr__(self, name: str) -> Any:
        return self.__values[name]



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="syncgit",
    version="0.2.4",
    author="RainingComputers",
    author_email="vishnu.vish.shankar@gmail.com",
    description="Sync python dicts, strings ad modules to git repository.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/RainingComputers/syncgit",
    packages=setuptools.find_packages(exclude=["docs", "tests"]),
    python_requires=">=3.8",
    install_requires=["PyYAML"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Topic :: Software Development :: Version Control :: Git",
        "Topic :: Utilities"
    ],
    keywords="syncgit"
)



from typing import Dict
from typing import Any

from threading import Lock


class ThreadSafeDict:
    def __init__(self) -> None:
        self.__values: Dict[str, Any] = {}
        self.__locks: Dict[str, Lock] = {}

    def __upsert_lock(self, key: str) -> Lock:
        try:
            lock = self.__locks[key]
        except KeyError:
            lock = Lock()
            self.__locks[key] = lock

        return lock

    def __setitem__(self, key: str, value: Any) -> Any:
        with self.__upsert_lock(key):
            self.__values[key] = value

    def __getitem__(self, key: str) -> Any:
        with self.__locks[key]:
            return self.__values[key]

    def __len__(self) -> int:
        return len(self.__values)
